// (C) 2001-2022 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


`timescale 1ps/1ps


module alt_cpriphy_c3_pat_gen_8b10b #(
    parameter DATA_WIDTH = 16
)(
    input             clk,
    input             reset,
    input [3:0]       cpri_rate_sel,
    output reg [DATA_WIDTH-1:0] tx_dataout,
    output reg [(DATA_WIDTH/8)-1:0] tx_ctrlout,
    output reg        start

);
localparam NUM_BYTES_PER_CYCLE = DATA_WIDTH/8;

// states
localparam SYNC_CTRL_WORD_1 = 4'h0;
localparam SYNC_CTRL_WORD_2 = 4'h1;
localparam SYNC_CTRL_WORD_3 = 4'h2;
localparam SYNC_CTRL_WORD_4 = 4'h3;
localparam SYNC_CTRL_WORD_5 = 4'h4;
localparam SYNC_CTRL_WORD_6 = 4'h5;
localparam SYNC_CTRL_WORD_7 = 4'h6;
localparam SYNC_CTRL_WORD_8 = 4'h7;
localparam DATA             = 4'h8;

// packet pattern
localparam SYNC_BYTE = 8'hBC; // K28.5
localparam D16_2 = 8'h50; // D16.2
localparam D5_6 = 8'hC5;  // D5.6
localparam FILLING_BYTE_1 = D16_2;
localparam FILLING_BYTE = D16_2;

///////////////////////////////////////////////////////////
//
// Hyperframe size (bytes) = X * W * T/8
//
// where X = number of basic frames per hyperframe = 256
//       W = number of words per basic frame = 16
//       T = length of word (bits) = 128 for 9.8G
//                                    80 for 6G
//                                    64 for 4.9G
//                                    40 for 3G
//                                    32 for 2.4G
//                                    16 for 1.2G
//       T/8 = length of word (bytes)
//
//////////////////////////////////////////////////////////////
wire [17:0] hyperframe_size = (cpri_rate_sel == 4'b0110) ? 18'd65536 : // 9.8G
                              (cpri_rate_sel == 4'b0101) ? 18'd40960 : // 6G
                              (cpri_rate_sel == 4'b0100) ? 18'd32768 : // 4.9G
                              (cpri_rate_sel == 4'b0011) ? 18'd20480 : // 3G
                              (cpri_rate_sel == 4'b0010) ? 18'd16384 : // 2.4G
                              (cpri_rate_sel == 4'b0001) ? 18'd8192  : // 1.2G
                                                           18'd65536;  // 9.8G

wire [16:0] hyperframe_cycle = hyperframe_size >> $clog2(NUM_BYTES_PER_CYCLE);

reg  [3:0]    state;
reg  [8:0]    tx_data_last; // 9 bit max count is 511

// Index for data cycle
reg [14:0]    frame_cycle;

// Generate Synchronization control word based on Table 9 of CPRI v7.0
always @ (posedge clk, posedge reset)
begin
    if (reset) begin
        state        <= SYNC_CTRL_WORD_1;
        tx_dataout   <= 16'h0000;
        tx_ctrlout   <= 2'b00;
        frame_cycle  <= 'd0;
        tx_data_last <= 'd0;
        start        <= 1'b0;
    end
    else begin
        case (state) // Based on Table 9 of CPRI v7.0
            //#Z.0.1, #Z.0.0
            SYNC_CTRL_WORD_1: begin
                tx_dataout  <= {FILLING_BYTE_1,SYNC_BYTE};
                tx_ctrlout  <= 2'b01;
                frame_cycle <= 'd0;
                tx_data_last <= 'd0;
                if (cpri_rate_sel == 4'b0001) // 1.2G
                    state <= DATA;
                else
                    state <= SYNC_CTRL_WORD_2;
            end

            //#Z.0.3, #Z.0.2
            SYNC_CTRL_WORD_2: begin
                tx_dataout  <= {FILLING_BYTE,FILLING_BYTE};
                tx_ctrlout  <= 2'b00;
                frame_cycle <= frame_cycle + 15'd1;
                tx_data_last <= 'd0;
                if (cpri_rate_sel == 4'b0010) // 2.4G
                    state <= DATA;
                else
                    state <= SYNC_CTRL_WORD_3;
            end

            //#Z.0.5, #Z.0.4
            SYNC_CTRL_WORD_3: begin
                tx_ctrlout  <= 2'b00;
                frame_cycle <= frame_cycle + 15'd1;
                tx_data_last <= 'd0;
                if (cpri_rate_sel == 4'b0011) begin // 3G
                    // 3G has weird situation where FILLING_BYTE ends at lower byte of tx_dataout (#Z.0.4)
                    // Pad upper byte of tx_dataout (#Z.0.5) with 0 (technically it is start of data byte).
                    tx_dataout  <= {8'h00,FILLING_BYTE};
                    state <= DATA;
                end else begin
                    tx_dataout  <= {FILLING_BYTE,FILLING_BYTE};
                    state <= SYNC_CTRL_WORD_4;
                end
            end

            //#Z.0.7, #Z.0.6
            SYNC_CTRL_WORD_4: begin
                tx_dataout  <= {FILLING_BYTE,FILLING_BYTE};
                tx_ctrlout  <= 2'b00;
                frame_cycle <= frame_cycle + 15'd1;
                tx_data_last <= 'd0;
                if (cpri_rate_sel == 4'b0100) // 4.9G
                    state <= DATA;
                else
                    state <= SYNC_CTRL_WORD_5;
            end

            //#Z.0.9, #Z.0.8
            SYNC_CTRL_WORD_5: begin
                tx_dataout  <= {FILLING_BYTE,FILLING_BYTE};
                tx_ctrlout  <= 2'b00;
                frame_cycle <= frame_cycle + 15'd1;
                tx_data_last <= 'd0;
                if (cpri_rate_sel == 4'b0101) // 6G
                    state <= DATA;
                else
                    state <= SYNC_CTRL_WORD_6;
            end

            //#Z.0.11, #Z.0.10
            SYNC_CTRL_WORD_6: begin
                tx_dataout  <= {FILLING_BYTE,FILLING_BYTE};
                tx_ctrlout  <= 2'b00;
                frame_cycle <= frame_cycle + 15'd1;
                tx_data_last <= 'd0;
                state <= SYNC_CTRL_WORD_7;
            end

            //#Z.0.13, #Z.0.12
            SYNC_CTRL_WORD_7: begin
                tx_dataout  <= {FILLING_BYTE,FILLING_BYTE};
                tx_ctrlout  <= 2'b00;
                frame_cycle <= frame_cycle + 15'd1;
                tx_data_last <= 'd0;
                state <= SYNC_CTRL_WORD_8;
            end

            //#Z.0.15, #Z.0.14
            SYNC_CTRL_WORD_8: begin
                tx_dataout  <= {FILLING_BYTE,FILLING_BYTE};
                tx_ctrlout  <= 2'b00;
                frame_cycle <= frame_cycle + 15'd1;
                tx_data_last <= 'd0;
                state <= DATA;
            end

            DATA : begin
                tx_ctrlout   <= 2'b00;
                tx_dataout   <= {tx_data_last[6:0],tx_data_last}; // Replicate data to pad dataout width.
                frame_cycle  <= frame_cycle + 15'd1;
                tx_data_last <= tx_data_last + 9'd1; // Generate data pattern (just a normal counter)

                if (frame_cycle < (hyperframe_cycle-2)) state <= DATA;
                else                                        state <= SYNC_CTRL_WORD_1;

                if (tx_data_last == 'd3) begin // cater for pipeline in checker
                    start <= 1'b1;
                end else if (tx_data_last >= 'd60) begin
                    start <= 1'b0;
                end else begin
                    start <= start;
                end
            end

            default: begin
                tx_dataout   <= 16'h0000;
                tx_ctrlout   <= 2'b00;
                tx_data_last <= 'd0;
                frame_cycle  <= 'd0;
                start        <= 1'b0;
            end
        endcase
    end
end
endmodule
