// (C) 2001-2022 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


`timescale 1ps/1ps


module alt_cpriphy_c3_pat_gen_xgmii
(
    input             clk,
    input             reset,
    input [3:0]       cpri_rate_sel,
    output reg [63:0] tx_dataout,
    output reg  [7:0] tx_ctrlout,
    output reg        start

);

// states
localparam SYNC_CTRL_WORD_1 = 2'h0;
localparam SYNC_CTRL_WORD_2 = 2'h1;
localparam DATA             = 2'h2;

// frame pattern
localparam SYNC_CTRL_WORD_D1 = 64'hfd50505050505050;
localparam SYNC_CTRL_WORD_C1 = 8'h80;
localparam SYNC_CTRL_WORD_D2 = 64'h50505050505050fb;
localparam SYNC_CTRL_WORD_C2 = 8'h01;
localparam DATA_C            = 8'h00;

///////////////////////////////////////////////////////////
//
// Hyperframe size (bytes) = X * W * T/8
//
// where X = number of basic frames per hyperframe = 256
//       W = number of words per basic frame = 16
//       T = length of word (bits) = 384 for 24G
//                                   192 for 12G
//                                   160 for 10G
//                                   128 for 8G
//       T/8 = length of word (bytes)
//
//////////////////////////////////////////////////////////////
wire [17:0] hyperframe_size = (cpri_rate_sel == 4'b1011) ? 18'd196608 : // 24G
                              (cpri_rate_sel == 4'b1010) ? 18'd98304 :  // 12G
                              (cpri_rate_sel == 4'b1001) ? 18'd81920 :  // 10G
                              (cpri_rate_sel == 4'b1000) ? 18'd65536 :  //  8G
                                                           18'd81920;   // 10G

reg  [1:0]    state;
reg  [8:0]    tx_data_last; // 9 bit max count is 511

// Index for XGMII cycle
reg [14:0]    xgmii_cycle;

always @ (posedge clk, posedge reset)
begin
    if (reset) begin
        state        <= SYNC_CTRL_WORD_1;
        tx_dataout   <= 64'h0000_0000_0000_0000;
        tx_ctrlout   <= 8'h00;
        xgmii_cycle  <= 'd0;
        tx_data_last <= 'd0;
        start        <= 1'b0;
    end
    else begin
        case (state)
            SYNC_CTRL_WORD_1: begin
                tx_dataout  <= SYNC_CTRL_WORD_D1;
                tx_ctrlout  <= SYNC_CTRL_WORD_C1;
                xgmii_cycle <= 'd0;
                tx_data_last <= 'd0;
                state       <= SYNC_CTRL_WORD_2;
            end

            SYNC_CTRL_WORD_2: begin
                tx_dataout  <= SYNC_CTRL_WORD_D2;
                tx_ctrlout  <= SYNC_CTRL_WORD_C2;
                state       <= DATA;
                xgmii_cycle <= xgmii_cycle + 15'd1;
                tx_data_last <= 'd0;
            end

            DATA : begin
                tx_ctrlout   <= DATA_C;
                tx_dataout   <= {tx_data_last[0],tx_data_last,tx_data_last,tx_data_last,tx_data_last,tx_data_last,tx_data_last,tx_data_last};
                xgmii_cycle  <= xgmii_cycle + 15'd1;
                tx_data_last <= tx_data_last + 9'd1;

                if (xgmii_cycle < ((hyperframe_size>>3)-2)) state <= DATA;
                else                                        state <= SYNC_CTRL_WORD_1;

                if (tx_data_last == 'd3) begin // cater for pipeline in checker
                    start <= 1'b1;
                end else if (tx_data_last >= 'd60) begin
                    start <= 1'b0;
                end else begin
                    start <= start;
                end
            end
            
            default: begin
                tx_dataout   <= 64'h0000_0000_0000_0000;
                tx_ctrlout   <= 8'h00;
                tx_data_last <= 'd0;
                xgmii_cycle  <= 'd0;
                start        <= 1'b0;
            end
        endcase
    end
end
endmodule
