// (C) 2001-2022 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


//-----------------------------------------------------------------------------
// Title         : pattern_checker
//-----------------------------------------------------------------------------
// File          :  Incremental data pattern checker
//-----------------------------------------------------------------------------



`timescale 1 ps/ 1 ps
module alt_cpriphy_c3_pat_chk_8b10b #(
    parameter DATA_WIDTH = 16
)(
    input  wire [DATA_WIDTH-1:0] datain,
    input  wire [(DATA_WIDTH/8)-1:0] ctrl,
    output reg hyperframe_sync,
    output reg  errorflag,
    output reg [4:0] err_cnt,
    input  wire clk,
    input  wire nreset,
    input [3:0] cpri_rate_sel,
    output reg stop
);

// states
localparam XACQ1   = 3'b000;
localparam XACQ2   = 3'b001;
localparam XSYNC1  = 3'b010;
localparam XSYNC2  = 3'b011;
localparam HFNSYNC = 3'b100;

// packet pattern
localparam SYNC_BYTE = 8'hBC; // K28.5
localparam D16_2 = 8'h50; // D16.2
localparam D5_6 = 8'hC5;  // D5.6
localparam FILLING_BYTE_1 = D16_2;
localparam FILLING_BYTE = D16_2;

///////////////////////////////////////////////////////////
//
// Hyperframe size (bytes) = X * W * T/8
//
// where X = number of basic frames per hyperframe = 256
//       W = number of words per basic frame = 16
//       T = length of word (bits) = 128 for 9.8G
//                                    80 for 6G
//                                    64 for 4.9G
//                                    40 for 3G
//                                    32 for 2.4G
//                                    16 for 1.2G
//       T/8 = length of word (bytes)
//
//////////////////////////////////////////////////////////////
wire [17:0] hyperframe_size = (cpri_rate_sel == 4'b0110) ? 18'd65536 : // 9.8G
                              (cpri_rate_sel == 4'b0101) ? 18'd40960 : // 6G
                              (cpri_rate_sel == 4'b0100) ? 18'd32768 : // 4.9G
                              (cpri_rate_sel == 4'b0011) ? 18'd20480 : // 3G
                              (cpri_rate_sel == 4'b0010) ? 18'd16384 : // 2.4G
                              (cpri_rate_sel == 4'b0001) ? 18'd8192  : // 1.2G
                                                           18'd65536;  // 9.8G

wire [16:0] hyperframe_cycle = hyperframe_size >> 1;

    wire cmp_data;
    wire is_sync_byte_cycle;
    reg [2:0] state;

    wire [DATA_WIDTH-1:0] datanext;
    reg [8:0] datanextreg; // 9 bit max count is 511.
                                 // The max count determines how far apart stop pulse is.
                                 // Max count must match that of pat gen (which generates start pulse).
    reg [DATA_WIDTH-1:0] datainreg, datainreg_p2;
    reg [7:0] ctrlreg, ctrlreg_p2;
    reg [16:0] frame_cycle;
    reg        is_data_cycle;

    always @ (posedge clk) begin
        datainreg <= datain;
        datainreg_p2 <= datainreg;
        ctrlreg   <= ctrl;
        ctrlreg_p2   <= ctrlreg;
    end

    assign is_sync_byte_cycle = (frame_cycle == 'd0); // Y=W=X=0

    // HFNSYNC detection based on Figure 26 of CPRI v7.0
    always @ (posedge clk or negedge nreset)
    begin
        if (~nreset) begin
            state         <= XACQ1;
        end
        else begin
        case (state)
            XACQ1:begin
                if ((datainreg[7:0] == SYNC_BYTE && ctrlreg[0] == 1'b1))
                    state <= XACQ2;
                else
                    state <= XACQ1;
            end
            
            XACQ2:begin
                if (is_sync_byte_cycle) begin
                    if ((datainreg[7:0] == SYNC_BYTE && ctrlreg[0] == 1'b1))
                        state <= XSYNC1;
                    else
                        state <= XACQ1;
                end else
                    state <= XACQ2;
            end
            
            XSYNC1:begin
                if (is_sync_byte_cycle) begin
                    if ((datainreg[7:0] == SYNC_BYTE && ctrlreg[0] == 1'b1))
                        state <= XSYNC2;
                    else
                        state <= XACQ1;
                end else
                    state <= XSYNC1;
            end
            
            XSYNC2:begin
                if (is_sync_byte_cycle) begin
                    if ((datainreg[7:0] == SYNC_BYTE && ctrlreg[0] == 1'b1))
                        state <= HFNSYNC;
                    else
                        state <= XSYNC1;
                end else
                    state <= XSYNC2;
            end

            HFNSYNC:begin
                if (is_sync_byte_cycle) begin
                    if ((datainreg[7:0] == SYNC_BYTE && ctrlreg[0] == 1'b1))
                        state <= HFNSYNC;
                    else
                        state <= XSYNC2;
                end else
                    state <= HFNSYNC;
            end
            
            default: begin
                state         <= XACQ1;
            end
        endcase // case (state)
    end
    end

    // Counter for length of frame, in terms of 16-bit cycle (based on datainreg)
    always @ (posedge clk or negedge nreset)
    begin
        if (~nreset) begin
            frame_cycle <= 'd0;
        end
        else begin
            if (state == XACQ1 && (datainreg[7:0] == SYNC_BYTE && ctrlreg[0] == 1'b1))
                frame_cycle <= 15'd1;
            else if (state != XACQ1 && frame_cycle < (hyperframe_cycle - 1))
                frame_cycle <= frame_cycle + 15'd1;
            else
                frame_cycle <= 15'd0;
        end
    end

    // Detect which 16-bit cycles contain data (based on datainreg_p2)
    always @ (posedge clk or negedge nreset)
    begin
        if (~nreset) begin
            is_data_cycle <= 1'b0;
        end
        else begin
            if ((cpri_rate_sel == 4'b0001 && frame_cycle >= 15'd1) || // For 1.2G, data starts at #Z.0.2, or cycle 1
                (cpri_rate_sel == 4'b0010 && frame_cycle >= 15'd2) || // For 2.4G, data starts at #Z.0.4, or cycle 2
                (cpri_rate_sel == 4'b0011 && frame_cycle >= 15'd3) || // For   3G, data starts at #Z.0.5, or cycle 3
                (cpri_rate_sel == 4'b0100 && frame_cycle >= 15'd4) || // For 4.9G, data starts at #Z.0.8, or cycle 4
                (cpri_rate_sel == 4'b0101 && frame_cycle >= 15'd5) || // For   6G, data starts at #Z.0.10, or cycle 5
                (cpri_rate_sel == 4'b0110 && frame_cycle >= 15'd8))   // For 9.8G, data starts at #Z.0.16, or cycle 8
                is_data_cycle <= 1'b1;
            else
                is_data_cycle <= 1'b0;
        end
    end

    assign cmp_data = ~(|ctrlreg_p2) && (state == HFNSYNC) && is_data_cycle;

    // Generate data pattern (just a normal counter)
    always @ (posedge clk or negedge nreset)
    begin
        if (~nreset) begin
            datanextreg <= 'd0;
            stop <= 1'b0;
            hyperframe_sync <= 1'b0;
        end
        else begin
            if (state == HFNSYNC)
                hyperframe_sync <= 1'b1;
            else
                hyperframe_sync <= 1'b0;
            
            if (cmp_data) begin
                datanextreg <= datanextreg + 9'b1;
                if (datanextreg == 'd0)
                    stop <= 1'b1;
                else if (datanextreg >=  'd59)
                    stop <= 1'b0;
                else
                    stop <= stop;
            end else begin
                datanextreg <= 'd0;
            end
        end
    end

    assign datanext = {datanextreg[6:0],datanextreg};

    // Compare received data (based on datainreg_p2) with data pattern
    always@(posedge clk or negedge nreset)begin
        if(~nreset)begin
            errorflag         <= 1'b0;
            err_cnt           <= 'd0;
        end
        else begin
            if((datanext != datainreg_p2) && cmp_data)begin
                err_cnt           <= err_cnt + 1'b1;
            end
            else begin
                err_cnt           <= err_cnt;
            end
            
            if (err_cnt == 'd0 && errorflag != 1'b1) begin
                errorflag         <= 1'b0;
            end else begin
                errorflag         <= 1'b1;
            end
        end
    end

endmodule
