// (C) 2001-2022 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


//-----------------------------------------------------------------------------
// Title         : pattern_checker
//-----------------------------------------------------------------------------
// File          :  Incremental data pattern checker
//-----------------------------------------------------------------------------



`timescale 1 ps/ 1 ps
module alt_cpriphy_c3_pat_chk_xgmii #(
    parameter NUM_CYCLES_FOR_LOCK = 8'd31,
    parameter INITIAL_VALUE = 97,
    parameter DATA_WIDTH = 8,
    parameter PRBS = 23,
    parameter PIPELINE = 3
)(
    input  wire [DATA_WIDTH-1:0] datain,
    input  wire [7:0] ctrl,
    output reg hyperframe_sync,
    output reg  errorflag,
    output reg [4:0] err_cnt,
    input  wire clk,
    input  wire nreset,
    input [3:0] cpri_rate_sel,
    output reg stop
);

// states
localparam XACQ1   = 3'b000;
localparam XACQ2   = 3'b001;
localparam XSYNC1  = 3'b010;
localparam XSYNC2  = 3'b011;
localparam HFNSYNC = 3'b100;

///////////////////////////////////////////////////////////
//
// Hyperframe size (bytes) = X * W * T/8
//
// where X = number of basic frames per hyperframe = 256
//       W = number of words per basic frame = 16
//       T = length of word (bits) = 384 for 24G
//                                   192 for 12G
//                                   160 for 10G
//                                   128 for 8G
//       T/8 = length of word (bytes)
//
//////////////////////////////////////////////////////////////
wire [17:0] hyperframe_size = (cpri_rate_sel == 4'b1011) ? 18'd196608 : // 24G
                              (cpri_rate_sel == 4'b1010) ? 18'd98304 :  // 12G
                              (cpri_rate_sel == 4'b1001) ? 18'd81920 :  // 10G
                              (cpri_rate_sel == 4'b1000) ? 18'd65536 :  //  8G
                                                           18'd81920;   // 10G

    wire cmp_data;
    reg [2:0] state;

    reg [DATA_WIDTH-1:0] datanext;
    reg [8:0] datanextreg; // 9 bit max count is 511
    reg [DATA_WIDTH-1:0] datainreg;
    reg [7:0] ctrlreg;
    reg [14:0] xgmii_cycle;

    always @ (posedge clk) begin
        datainreg <= datain;
        ctrlreg   <= ctrl;
    end

    assign cmp_data = ~(|ctrlreg);

    always @ (posedge clk or negedge nreset)
    begin
        if (~nreset) begin
            state         <= XACQ1;
        end
        else begin
        case (state)
            XACQ1:begin
                if ((datainreg[7:0] == 8'hfb && ctrlreg[0] == 1'b1))
                    state <= XACQ2;
                else
                    state <= XACQ1;
            end
            
            XACQ2:begin
                if (xgmii_cycle == 'd1) begin
                    if ((datainreg[7:0] == 8'hfb && ctrlreg[0] == 1'b1))
                        state <= XSYNC1;
                    else
                        state <= XACQ1;
                end else
                    state <= XACQ2;
            end
            
            XSYNC1:begin
                if (xgmii_cycle == 'd1) begin
                    if ((datainreg[7:0] == 8'hfb && ctrlreg[0] == 1'b1))
                        state <= XSYNC2;
                    else
                        state <= XACQ1;
                end else
                    state <= XSYNC1;
            end
            
            XSYNC2:begin
                if (xgmii_cycle == 'd1) begin
                    if ((datainreg[7:0] == 8'hfb && ctrlreg[0] == 1'b1))
                        state <= HFNSYNC;
                    else
                        state <= XSYNC1;
                end else
                    state <= XSYNC2;
            end

            HFNSYNC:begin
                if (xgmii_cycle == 'd1) begin
                    if ((datainreg[7:0] == 8'hfb && ctrlreg[0] == 1'b1))
                        state <= HFNSYNC;
                    else
                        state <= XSYNC2;
                end else
                    state <= HFNSYNC;
            end
            
            default: begin
                state         <= XACQ1;
            end
        endcase // case (state)
    end
    end

    always @ (posedge clk or negedge nreset)
    begin
        if (~nreset) begin
            xgmii_cycle <= 'd0;
        end
        else begin
            if (state == XACQ1 && (datainreg[7:0] == 8'hfb && ctrlreg[0] == 1'b1))
                xgmii_cycle <= 15'd2;
            else if (state != XACQ1 && xgmii_cycle < ((hyperframe_size>>3)-1))
                xgmii_cycle <= xgmii_cycle + 15'd1;
            else
                xgmii_cycle <= 15'd0;
        end
    end

    always @ (posedge clk or negedge nreset)
    begin
        if (~nreset) begin
            datanextreg <= 'd0;
            stop <= 1'b0;
            hyperframe_sync <= 1'b0;
        end
        else begin
            if (state == HFNSYNC)
                hyperframe_sync <= 1'b1;
            else
                hyperframe_sync <= 1'b0;
            
            if (state == HFNSYNC && cmp_data) begin
                datanextreg <= datanextreg + 9'b1;
                if (datanextreg == 'd1)
                    stop <= 1'b1;
                else if (datanextreg >=  'd60)
                    stop <= 1'b0;
                else
                    stop <= stop;
            end else begin
                datanextreg <= 'd0;
            end
        end
    end

    always@(posedge clk or negedge nreset) begin
        if (~nreset) begin
            errorflag <= 1'b0;
            err_cnt   <= 'd0;
        end
        else begin
            if (((datanextreg != datainreg[8:0])   || 
                (datanextreg != datainreg[17:9])  || 
                (datanextreg != datainreg[26:18]) || 
                (datanextreg != datainreg[35:27]) || 
                (datanextreg != datainreg[44:36]) || 
                (datanextreg != datainreg[53:45]) || 
                (datanextreg != datainreg[62:54]) || 
                (datanextreg[0] != datainreg[63])) && (cmp_data && (state == HFNSYNC))) begin
                    err_cnt <= err_cnt + 1'b1;
            end
            else begin
                err_cnt <= err_cnt;
            end

            if (err_cnt == 'd0 && errorflag != 1'b1) begin
                errorflag   <= 1'b0;
            end
            else begin
                errorflag   <= 1'b1;
            end
        end
    end

endmodule
